@@ -1,7 +1,7 @@
-/* @(#)fifo.c	1.28 01/04/11 Copyright 1989,1997 J. Schilling */
+/* @(#)fifo.c	1.47 04/03/02 Copyright 1989,1997-2004 J. Schilling */
 #ifndef lint
 static	char sccsid[] =
-	"@(#)fifo.c	1.28 01/04/11 Copyright 1989,1997 J. Schilling";
+	"@(#)fifo.c	1.47 04/03/02 Copyright 1989,1997-2004 J. Schilling";
 #endif
 /*
  *	A "fifo" that uses shared memory between two processes
@@ -10,30 +10,34 @@ static	char sccsid[] =
  *	and a proposal from Finn Arne Gangstad <finnag@guardian.no>
  *	who had the idea to use a ring buffer to handle average size chunks.
  *
- *	Copyright (c) 1989,1997 J. Schilling
+ *	Copyright (c) 1989,1997-2004 J. Schilling
  */
 /*
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2, or (at your option)
- * any later version.
+ * it under the terms of the GNU General Public License version 2
+ * as published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; see the file COPYING.  If not, write to
- * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
+ * You should have received a copy of the GNU General Public License along with
+ * this program; see the file COPYING.  If not, write to the Free Software
+ * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+#ifndef	DEBUG
 #define	DEBUG
+#endif
 /*#define	XDEBUG*/
 #include <mconfig.h>
 #if	!defined(HAVE_SMMAP) && !defined(HAVE_USGSHM) && !defined(HAVE_DOSALLOCSHAREDMEM)
 #undef	FIFO			/* We cannot have a FIFO on this platform */
 #endif
+#if	!defined(HAVE_FORK)
+#undef	FIFO			/* We cannot have a FIFO on this platform */
+#endif
 #ifdef	FIFO
 #if !defined(USE_MMAP) && !defined(USE_USGSHM)
 #define	USE_MMAP
@@ -52,16 +56,15 @@ static	char sccsid[] =
 #	define	USE_OS2SHM
 #endif
 
+#include <stdio.h>
+#include <stdxlib.h>
+#include <unixstd.h>	/* includes <sys/types.h> */
+#include <utypes.h>
 #include <fctldefs.h>
-#include <sys/types.h>
 #if defined(HAVE_SMMAP) && defined(USE_MMAP)
 #include <mmapdefs.h>
 #endif
-#include <stdio.h>
-#include <stdxlib.h>
-#include <unixstd.h>
 #include <waitdefs.h>
-#include <utypes.h>
 #include <standard.h>
 #include <errno.h>
 #include <signal.h>
@@ -69,12 +72,13 @@ static	char sccsid[] =
 #include <schily.h>
 
 #include "cdrecord.h"
+#include "xio.h"
 
 #ifdef DEBUG
 #ifdef XDEBUG
 FILE	*ef;
-#define	USDEBUG1	if (debug) {if(s == owner_reader)fprintf(ef, "r");else fprintf(ef, "w");fflush(ef);}
-#define	USDEBUG2	if (debug) {if(s == owner_reader)fprintf(ef, "R");else fprintf(ef, "W");fflush(ef);}
+#define	USDEBUG1	if (debug) {if (s == owner_reader) fprintf(ef, "r"); else fprintf(ef, "w"); fflush(ef); }
+#define	USDEBUG2	if (debug) {if (s == owner_reader) fprintf(ef, "R"); else fprintf(ef, "W"); fflush(ef); }
 #else
 #define	USDEBUG1
 #define	USDEBUG2
@@ -86,13 +90,13 @@ FILE	*ef;
 #define	USDEBUG2
 #endif
 
-#define palign(x, a)	(((char *)(x)) + ((a) - 1 - (((unsigned)((x)-1))%(a))))
+#define	palign(x, a)	(((char *)(x)) + ((a) - 1 - (((UIntptr_t)((x)-1))%(a))))
 
 typedef enum faio_owner {
-	owner_none,		/* Unused in real life			     */
-	owner_writer,		/* owned by process that writes into FIFO    */
-	owner_faio,		/* Intermediate state when buf still in use  */
-	owner_reader		/* owned by process that reads from FIFO     */
+	owner_none,		/* Unused in real life			    */
+	owner_writer,		/* owned by process that writes into FIFO   */
+	owner_faio,		/* Intermediate state when buf still in use */
+	owner_reader		/* owned by process that reads from FIFO    */
 } fowner_t;
 
 char	*onames[] = {
@@ -142,32 +146,32 @@ struct faio_stats {
 LOCAL	char	*buf;
 LOCAL	char	*bufbase;
 LOCAL	char	*bufend;
-LOCAL	long	buflen;
+LOCAL	long	buflen;			/* The size of the FIFO buffer */
 
 extern	int	debug;
 extern	int	lverbose;
 
 EXPORT	void	init_fifo	__PR((long));
 #ifdef	USE_MMAP
-LOCAL	char*	mkshare		__PR((int size));
+LOCAL	char	*mkshare	__PR((int size));
 #endif
 #ifdef	USE_USGSHM
-LOCAL	char*	mkshm		__PR((int size));
+LOCAL	char	*mkshm		__PR((int size));
 #endif
 #ifdef	USE_OS2SHM
-LOCAL	char*	mkos2shm	__PR((int size));
+LOCAL	char	*mkos2shm	__PR((int size));
 #endif
 
-EXPORT	BOOL	init_faio	__PR((int tracks, track_t *track, int));
+EXPORT	BOOL	init_faio	__PR((track_t *trackp, int));
 EXPORT	BOOL	await_faio	__PR((void));
 EXPORT	void	kill_faio	__PR((void));
 EXPORT	int	wait_faio	__PR((void));
-LOCAL	void	faio_reader	__PR((int tracks, track_t *track));
+LOCAL	void	faio_reader	__PR((track_t *trackp));
 LOCAL	void	faio_read_track	__PR((track_t *trackp));
 LOCAL	void	faio_wait_on_buffer __PR((faio_t *f, fowner_t s,
-					  unsigned long delay,
-					  unsigned long max_wait));
-LOCAL	int	faio_read_segment __PR((int fd, faio_t *f, int len));
+					unsigned long delay,
+					unsigned long max_wait));
+LOCAL	int	faio_read_segment __PR((int fd, faio_t *f, track_t *track, long secno, int len));
 LOCAL	faio_t	*faio_ref	__PR((int n));
 EXPORT	int	faio_read_buf	__PR((int f, char *bp, int size));
 EXPORT	int	faio_get_buf	__PR((int f, char **bpp, int size));
@@ -241,9 +245,11 @@ mkshare(size)
 #endif
 	if (addr == (char *)-1)
 		comerr("Cannot get mmap for %d Bytes on /dev/zero.\n", size);
-	close(f);
+	if (f >= 0)
+		close(f);
 
-	if (debug) errmsgno(EX_BAD, "shared memory segment attached: %p\n", (void *)addr);
+	if (debug) errmsgno(EX_BAD, "shared memory segment attached at: %p size %d\n",
+				(void *)addr, size);
 
 	return (addr);
 }
@@ -280,7 +286,8 @@ mkshm(size)
 	if ((addr = shmat(id, (char *)0, 0600)) == (char *)-1)
 		comerr("shmat failed\n");
 
-	if (debug) errmsgno(EX_BAD, "shared memory segment attached: %p\n", addr);
+	if (debug) errmsgno(EX_BAD, "shared memory segment attached at: %p size %d\n",
+				(void *)addr, size);
 
 	if (shmctl(id, IPC_RMID, 0) < 0)
 		comerr("shmctl failed to detach shared memory segment\n");
@@ -309,12 +316,12 @@ mkos2shm(size)
 	 * The OS/2 implementation of shm (using shm.dll) limits the size of one shared
 	 * memory segment to 0x3fa000 (aprox. 4MBytes). Using OS/2 native API we have
 	 * no such restriction so I decided to use it allowing fifos of arbitrary size.
-         */
-	if(DosAllocSharedMem(&addr,NULL,size,0X100L | 0x1L | 0x2L | 0x10L))
+	 */
+	if (DosAllocSharedMem(&addr, NULL, size, 0X100L | 0x1L | 0x2L | 0x10L))
 		comerr("DosAllocSharedMem() failed\n");
 
-	if (debug)
-		errmsgno(EX_BAD, "shared memory allocated at address: %x\n", addr);
+	if (debug) errmsgno(EX_BAD, "shared memory allocated attached at: %p size %d\n",
+				(void *)addr, size);
 
 	return (addr);
 }
@@ -322,18 +329,34 @@ mkos2shm(size)
 
 LOCAL	int	faio_buffers;
 LOCAL	int	faio_buf_size;
-LOCAL	int	buf_idx;
+LOCAL	int	buf_idx = 0;		/* Initialize to fix an Amiga bug   */
+LOCAL	int	buf_idx_reader = 0;	/* Separate var to allow vfork()    */
+					/* buf_idx_reader is for the process */
+					/* that fills the FIFO		    */
 LOCAL	pid_t	faio_pid;
 LOCAL	BOOL	faio_didwait;
 
+#ifdef AMIGA
+/*
+ * On Amiga fork will be replaced by the speciall vfork() like call ix_vfork,
+ * which lets the parent asleep. The child process later wakes up the parent
+ * process by calling ix_fork_resume().
+ */
+#define	fork()		 ix_vfork()
+#define	__vfork_resume() ix_vfork_resume()
+
+#else	/* !AMIGA */
+#define	__vfork_resume()
+#endif
+
+
 /*#define	faio_ref(n)	(&((faio_t *)buf)[n])*/
 
 
 EXPORT BOOL
-init_faio(tracks, track, bufsize)
-	int	tracks;
-	track_t	*track;
-	int	bufsize;
+init_faio(trackp, bufsize)
+	track_t	*trackp;
+	int	bufsize;	/* The size of a single transfer buffer */
 {
 	int	n;
 	faio_t	*f;
@@ -358,16 +381,16 @@ init_faio(tracks, track, bufsize)
 	 * properly page aligned.
 	 */
 	bufsize = roundup(bufsize, pagesize);
-	faio_buffers = (buflen - sizeof(*sp)) / bufsize;
-	EDEBUG(("bufsize: %d buffers: %d hdrsize %ld\n", bufsize, faio_buffers, (long)faio_buffers * sizeof(struct faio)));
+	faio_buffers = (buflen - sizeof (*sp)) / bufsize;
+	EDEBUG(("bufsize: %d buffers: %d hdrsize %ld\n", bufsize, faio_buffers, (long)faio_buffers * sizeof (struct faio)));
 
 	/*
 	 * Reduce buffer space by header space.
 	 */
-	n = sizeof(*sp) + faio_buffers * sizeof(struct faio);
+	n = sizeof (*sp) + faio_buffers * sizeof (struct faio);
 	n = roundup(n, pagesize);
 	faio_buffers = (buflen-n) / bufsize;
-	EDEBUG(("bufsize: %d buffers: %d hdrsize %ld\n", bufsize, faio_buffers, (long)faio_buffers * sizeof(struct faio)));
+	EDEBUG(("bufsize: %d buffers: %d hdrsize %ld\n", bufsize, faio_buffers, (long)faio_buffers * sizeof (struct faio)));
 
 	if (faio_buffers < MIN_BUFFERS) {
 		errmsgno(EX_BAD,
@@ -380,7 +403,7 @@ init_faio(tracks, track, bufsize)
 		printf("Using %d buffers of %d bytes.\n", faio_buffers, faio_buf_size);
 
 	f = (faio_t *)buf;
-	base = buf + roundup(sizeof(*sp) + faio_buffers * sizeof(struct faio),
+	base = buf + roundup(sizeof (*sp) + faio_buffers * sizeof (struct faio),
 				pagesize);
 
 	for (n = 0; n < faio_buffers; n++, f++, base += bufsize) {
@@ -405,18 +428,33 @@ init_faio(tracks, track, bufsize)
 		raisepri(1);		/* almost max priority */
 
 #ifdef USE_OS2SHM
-		DosGetSharedMem(buf,3);	/* PAG_READ|PAG_WRITE */
+		DosGetSharedMem(buf, 3); /* PAG_READ|PAG_WRITE */
 #endif
 		/* Ignoring SIGALRM cures the SCO usleep() bug */
 /*		signal(SIGALRM, SIG_IGN);*/
-		faio_reader(tracks, track);
+		__vfork_resume();	/* Needed on some platforms */
+		faio_reader(trackp);
 		/* NOTREACHED */
 	} else {
+#ifdef	__needed__
+		Uint	t;
+#endif
+
 		faio_didwait = FALSE;
 
+		/*
+		 * XXX We used to close all track files in the foreground
+		 * XXX process. This was not correct before we used "xio"
+		 * XXX and with "xio" it will start to fail because we need
+		 * XXX the fd handles for the faio_get_buf() function.
+		 */
+#ifdef	__needed__
 		/* close all file-descriptors that only the child will use */
-		for (n = 1; n <= tracks; n++)
-			close(track[n].f);
+		for (t = 1; t <= trackp->tracks; t++) {
+			if (trackp[t].xfp != NULL)
+				xclose(trackp[t].xfp);
+		}
+#endif
 	}
 
 	return (TRUE);
@@ -476,20 +514,19 @@ wait_faio()
 }
 
 LOCAL void
-faio_reader(tracks, track)
-	int	tracks;
-	track_t	*track;
+faio_reader(trackp)
+	track_t	*trackp;
 {
 	/* This function should not return, but _exit. */
-	int	trackno;
+	Uint	trackno;
 
 	if (debug)
 		printf("\nfaio_reader starting\n");
 
-	for (trackno = 1; trackno <= tracks; trackno++) {
+	for (trackno = 1; trackno <= trackp->tracks; trackno++) {
 		if (debug)
-			printf("\nfaio_reader reading track %d\n", trackno);
-		faio_read_track(&track[trackno]);
+			printf("\nfaio_reader reading track %u\n", trackno);
+		faio_read_track(&trackp[trackno]);
 	}
 	sp->done++;
 	if (debug)
@@ -523,45 +560,58 @@ LOCAL void
 faio_read_track(trackp)
 	track_t *trackp;
 {
-	int	fd = trackp->f;
+	int	fd = -1;
 	int	bytespt = trackp->secsize * trackp->secspt;
+	int	secspt = trackp->secspt;
 	int	l;
-	off_t	tracksize = trackp->tracksize;
-	off_t	bytes_read = (off_t)0;
+	long	secno = trackp->trackstart;
+	tsize_t	tracksize = trackp->tracksize;
+	tsize_t	bytes_read = (tsize_t)0;
 	long	bytes_to_read;
 
+	if (trackp->xfp != NULL)
+		fd = xfileno(trackp->xfp);
+
 	if (bytespt > faio_buf_size) {
 		comerrno(EX_BAD,
 		"faio_read_track fatal: secsize %d secspt %d, bytespt(%d) > %d !!\n",
-			 trackp->secsize, trackp->secspt, bytespt,
-			 faio_buf_size);
+			trackp->secsize, trackp->secspt, bytespt,
+			faio_buf_size);
 	}
 
 	do {
 		bytes_to_read = bytespt;
 		if (tracksize > 0) {
-			if ((tracksize - bytes_read) > bytespt)
+			if ((tracksize - bytes_read) > bytespt) {
 				bytes_to_read = bytespt;
-			else
-				bytes_to_read = tracksize - bytes_read;				
+			} else {
+				bytes_to_read = tracksize - bytes_read;
+			}
 		}
-		l = faio_read_segment(fd, faio_ref(buf_idx), bytes_to_read);
-		if (++buf_idx >= faio_buffers)
-			buf_idx = 0;
+		l = faio_read_segment(fd, faio_ref(buf_idx_reader), trackp, secno, bytes_to_read);
+		if (++buf_idx_reader >= faio_buffers)
+			buf_idx_reader = 0;
 		if (l <= 0)
 			break;
 		bytes_read += l;
+		secno += secspt;
 	} while (tracksize < 0 || bytes_read < tracksize);
 
-	close(fd);	/* Don't keep files open longer than neccesary */
+	xclose(trackp->xfp);	/* Don't keep files open longer than neccesary */
 }
 
 LOCAL void
+#ifdef	PROTOTYPES
+faio_wait_on_buffer(faio_t *f, fowner_t s,
+			unsigned long delay,
+			unsigned long max_wait)
+#else
 faio_wait_on_buffer(f, s, delay, max_wait)
 	faio_t	*f;
 	fowner_t s;
 	unsigned long delay;
 	unsigned long max_wait;
+#endif
 {
 	unsigned long max_loops;
 
@@ -586,16 +636,18 @@ faio_wait_on_buffer(f, s, delay, max_wait)
 	if (debug) {
 		errmsgno(EX_BAD,
 		"%lu microseconds passed waiting for %d current: %d idx: %ld\n",
-		max_wait, s, f->owner, (long)(f - faio_ref(0))/sizeof(*f));
+		max_wait, s, f->owner, (long)(f - faio_ref(0))/sizeof (*f));
 	}
 	comerrno(EX_BAD, "faio_wait_on_buffer for %s timed out.\n",
 	(s > owner_reader || s < owner_none) ? "bad_owner" : onames[s-owner_none]);
 }
 
 LOCAL int
-faio_read_segment(fd, f, len)
+faio_read_segment(fd, f, trackp, secno, len)
 	int	fd;
 	faio_t	*f;
+	track_t	*trackp;
+	long	secno;
 	int	len;
 {
 	int l;
@@ -603,15 +655,15 @@ faio_read_segment(fd, f, len)
 	faio_wait_on_buffer(f, owner_writer, WRITER_DELAY, WRITER_MAXWAIT);
 
 	f->fd = fd;
-	l = read_buf(fd, f->bufp, len);
+	l = fill_buf(fd, trackp, secno, f->bufp, len);
 	f->len = l;
-	f->saved_errno = errno;
+	f->saved_errno = geterrno();
 	f->owner = owner_reader;
 	f->users = sp->users;
 
 	sp->puts++;
 
-	return l;
+	return (l);
 }
 
 EXPORT int
@@ -626,7 +678,7 @@ faio_read_buf(fd, bp, size)
 	if (len > 0) {
 		movebytes(bufp, bp, len);
 	}
-	return len;
+	return (len);
 }
 
 EXPORT int
@@ -653,7 +705,7 @@ again:
 	}
 	faio_wait_on_buffer(f, owner_reader, READER_DELAY, READER_MAXWAIT);
 	len = f->len;
-	
+
 	if (f->fd != fd) {
 		if (f->len == 0) {
 			/*
@@ -673,14 +725,14 @@ again:
 	}
 
 	if (len < 0)
-		errno = f->saved_errno;
+		seterrno(f->saved_errno);
 
 	sp->gets++;
 
 	*bpp = f->bufp;
 	if (--f->users <= 0)
 		f->owner = owner_faio;
-	return len;
+	return (len);
 }
 
 EXPORT void
@@ -701,8 +753,9 @@ fifo_percent(addone)
 {
 	int	percent;
 
-	if (buflen == 0L)
+	if (sp == NULL)	/* We might not use a FIFO */
 		return (-1);
+
 	if (sp->done)
 		return (100);
 	percent = (100*(sp->puts + 1 - sp->gets)/faio_buffers);
@@ -713,13 +766,13 @@ fifo_percent(addone)
 #else	/* FIFO */
 
 #include <standard.h>
-#include <sys/types.h>
-#include <utypes.h>
+#include <utypes.h>	/* includes sys/types.h */
+#include <schily.h>
 
 #include "cdrecord.h"
 
 EXPORT	void	init_fifo	__PR((long));
-EXPORT	BOOL	init_faio	__PR((int tracks, track_t *track, int));
+EXPORT	BOOL	init_faio	__PR((track_t *track, int));
 EXPORT	BOOL	await_faio	__PR((void));
 EXPORT	void	kill_faio	__PR((void));
 EXPORT	int	wait_faio	__PR((void));
@@ -737,8 +790,7 @@ init_fifo(fs)
 }
 
 EXPORT BOOL
-init_faio(tracks, track, bufsize)
-	int	tracks;
+init_faio(track, bufsize)
 	track_t	*track;
 	int	bufsize;
 {
