@@ -1,6 +1,6 @@
-/* @(#)fdup.c	1.12 01/12/09 Copyright 1986, 1995 J. Schilling */
+/* @(#)fdup.c	1.14 04/08/08 Copyright 1986, 1995-2003 J. Schilling */
 /*
- *	Copyright (c) 1986, 1995 J. Schilling
+ *	Copyright (c) 1986, 1995-2003 J. Schilling
  */
 /*
  * This program is free software; you can redistribute it and/or modify
@@ -13,15 +13,15 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; see the file COPYING.  If not, write to
- * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
+ * You should have received a copy of the GNU General Public License along with
+ * this program; see the file COPYING.  If not, write to the Free Software
+ * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-#include "io.h"
+#include "schilyio.h"
 
 /*
- * Note that because of a definition in io.h we are using fseeko()/ftello()
+ * Note that because of a definition in schilyio.h we are using fseeko()/ftello()
  * instead of fseek()/ftell() if available.
  */
 
@@ -33,9 +33,8 @@ fdup(f)
 
 	down(f);
 	if ((newfd = dup(fileno(f))) < 0)
-        	return (FILE *) NULL;
+		return ((FILE *) NULL);
 
 	lseek(newfd, ftell(f), SEEK_SET);
 	return (_fcons((FILE *)0, newfd, (FI_READ | FI_WRITE | FI_CLOSE)));
-} 
-
+}
